;;; MD5 hash implementation
;;; RFC 1321 - The MD5 Message-Digest Algorithm
;;; (c) Ufasoft 2011
;;; This software is PUBLIC DOMAIN

(require :hash-common)

(defpackage :md5
  (:export #:md5)
  (:use :cl :hash-common))

(in-package :md5)

(defconstant k (let ((v (make-array 64)))
                 (dotimes (i 64 v)
                   (setf (svref v i) (floor (scale-float (abs (sin (float (1+ i) 1.0d0))) 32))))))

(defconstant g (let ((v (make-array 64)))
                 (dotimes (i 64 v)
                   (setf (svref v i) (mod (cond ((<= i 15) i)
                                                ((<= i 31) (+ (* 5 i) 1))
                                                ((<= i 47) (+ (* 3 i) 5))
                                                (t (* 7 i)))
                                          16)))))

(defconstant r (let ((v (make-array 64)))
                 (dotimes (i 64 v)
                   (setf (svref v i) (svref (cond ((<= i 15) #(7 12 17 22))
                                                  ((<= i 31) #(5 9 14 20))
                                                  ((<= i 47) #(4 11 16 23))
                                                  (t #(6 10 15 21)))
                                            (logand i 3))))))

(defconstant aaa #x67452301)
(defconstant bbb #xefcdab89)
(defconstant ccc #x98badcfe)
(defconstant ddd #x10325476)



(defun f (x y z)
  (logior (logand x y) (logandc1 x z)))

(defun g (x y z)
  (logior (logand x z) (logandc2 y z)))

(defun h (x y z)
  (logxor x y z))

(defun i (x y z)
  (logxor y (logorc2 x z)))

(defconstant funs (vector #'f #'g #'h #'i))


(defun <<< (x s)
  (logior (ldb (byte s (- 32 s)) x)
          (ash (ldb (byte (- 32 s) 0) x) s)))

(defun swab (x)
  (logior (ash (ldb (byte 8 0) x) 24)
          (ash (ldb (byte 8 8) x) 16)
          (ash (ldb (byte 8 16) x) 8)
          (ldb (byte 8 24) x)))


(defun init-hasher ()
  (let* ((aa aaa)
         (bb bbb)
         (cc ccc)
         (dd ddd)
         (a aa)
         (b bb)
         (c cc)
         (d dd)
         (m (make-array 16)))
    (values (lambda (blk)
                (dotimes (i 16)
                  (let ((n (ash i 2)))
                    (setf (svref m i) (logior (svref blk n)
                                            (ash (svref blk (+ n 1)) 8)
                                            (ash (svref blk (+ n 2)) 16)
                                            (ash (svref blk (+ n 3)) 24)))))
                (dotimes (i 64)
                  (setq a (ldb (byte 32 0) (+ b (<<< (+ a
                                                        (funcall (svref funs (ash i -4)) b c d)
                                                        (svref m (svref g i))
                                                        (svref k i))
                                                     (svref r i)))))
                  (rotatef a d c b))
                (setq aa (ldb (byte 32 0) (+ aa a))
                      bb (ldb (byte 32 0) (+ bb b))
                      cc (ldb (byte 32 0) (+ cc c))
                      dd (ldb (byte 32 0) (+ dd d))))
            (lambda () 
              (logior (ash (swab aa) 96) (ash (swab bb) 64) (ash (swab cc) 32) (swab dd))))))
             


(defun md5 (stm)
    "Calculates MD5 hash of (unsigned-byte 8) stream"
  (multiple-value-call #'hash-common stm (init-hasher)))
